"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2020-now @ Copyright Wechaty

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import re
from typing import (
    Optional,
)

from wechaty_puppet import (
    FileBox,
    get_logger
)


log = get_logger('Config')

# log.debug('test logging debug')
# log.info('test logging info')


# TODO(wj-Mcat): there is no reference usage, so need to be removed
_FILE_PATH = os.path.dirname(os.path.realpath(__file__))
DATA_PATH = os.path.realpath(
    os.path.join(
        _FILE_PATH,
        '../data',
    ),
)

# http://jkorpela.fi/chars/spaces.html
# String.fromCharCode(8197)
AT_SEPARATOR = chr(0x2005)

# refer to:https://github.com/wechaty/python-wechaty/issues/285#issuecomment-997441596
PARALLEL_TASK_NUM = 100


def global_exception_handler(exception: Exception) -> None:
    """
    handle the global exception
    :param exception: exception message
    :return:
    """
    log.error('occur %s %s', exception.__class__.__name__, str(exception.args))
    print(exception)


class DefaultSetting(dict):
    """
    store global default setting
    """
    default_api_host: Optional[str] = None
    default_port: Optional[int] = None
    default_protocol: Optional[str] = None


# pylint: disable=R0903
def valid_api_host(api_host: str) -> bool:
    """
    test the validation of the api_host
    :param api_host:
    :return:
    """
    pattern = re.compile(
        r'^(([a-zA-Z]{1})|([a-zA-Z]{1}[a-zA-Z]{1})|:?[0-9]*'
        r'([a-zA-Z]{1}[0-9]{1})|([0-9]{1}[a-zA-Z]{1})|:?[0-9]*'
        r'([a-zA-Z0-9][-_.a-zA-Z0-9]{0,61}[a-zA-Z0-9]))\.:?[0-9]*'
        r'([a-zA-Z]{2,13}|[a-zA-Z0-9-]{2,30}.[a-zA-Z]{2,3}):?[0-9]*$'
    )
    return bool(pattern.match(api_host))


class Config:
    """
    get the configuration from the environment variables
    """
    @property
    def cache_dir(self) -> str:
        """get the cache dir in the lazy loading mode

        Returns:
            str: the path of cache dir
        """
        return os.environ.get("CACHE_DIR", '.wechaty')

    @property
    def ui_dir(self) -> str:
        """get the ui directory

        Returns:
            str: the path of the ui dir
        """
        default_ui_dir = os.path.join(
            os.path.dirname(__file__),
            'ui'
        )
        return os.environ.get("UI_DIR", default_ui_dir)


# export const CHATIE_OFFICIAL_ACCOUNT_ID = 'gh_051c89260e5d'
# chatie_official_account_id = 'gh_051c89260e5d'
# CHATIE_OFFICIAL_ACCOUNT_ID = 'gh_051c89260e5d'


def qr_code_for_chatie() -> FileBox:
    """
    create QRcode for chatie
    :return:
    """
    # const CHATIE_OFFICIAL_ACCOUNT_QRCODE =
    # 'http://weixin.qq.com/r/qymXj7DEO_1ErfTs93y5'
    chatie_official_account_qr_code: str = \
        'http://weixin.qq.com/r/qymXj7DEO_1ErfTs93y5'
    return FileBox.from_qr_code(chatie_official_account_qr_code)


config = Config()



"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2020-now @ Copyright Wechaty

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

from typing import List, Optional, Any
import asyncio

from wechaty_puppet.file_box import FileBox

from wechaty_puppet.schemas.message import (
    MessageQueryFilter,
)
from wechaty_puppet.schemas.mini_program import (
    MiniProgramPayload
)
# pylint: disable=R0904
from wechaty_puppet import Puppet
from wechaty_puppet.schemas.puppet import PuppetOptions
from wechaty_puppet.schemas.event import EventLoginPayload
from wechaty_puppet.schemas.room import (
    RoomQueryFilter
)
from wechaty_puppet.schemas.url_link import UrlLinkPayload

from wechaty_puppet.schemas.types import (
    FriendshipPayload,
    ContactPayload,
    ImageType,
    MessagePayload,

    RoomPayload,
    RoomMemberPayload,

    RoomInvitationPayload,

    PayloadType

)
from pyee import AsyncIOEventEmitter


class FakePuppet(Puppet):
    """
    puppet interface class, which is the abstract of puppet implementation.

    python-wechaty-puppet-XX can be all of the protocol of IM software.
    """

    def __init__(self, options: PuppetOptions, name: str = 'puppet') -> None:
        super().__init__(options, name)
        self.name: str = name
        self.options = options
        self.emitter = AsyncIOEventEmitter()

    async def message_image(
            self,
            message_id: str,
            image_type: ImageType
    ) -> FileBox:
        """
        docstring
        :param message_id:
        :param image_type:
        :return:
        """

    async def ding(self, data: Optional[str] = None) -> None:
        """
        set the ding event
        :param data:
        :return:
        """

    def on(self, event_name: str, caller: Any) -> None:
        """register the event

        Args:
            event_name (str): the name of event
            caller (Any): the target call function
        """
        self.emitter.on(event_name, caller)

    def listener_count(self, event_name: str) -> int:
        """
        get the count of a specific event listener
        """

    async def start(self) -> None:
        """
        start the puppet
        :return:
        """
        payload = EventLoginPayload(contact_id='fake-contact-id')

        self.emitter.emit('login', payload)
        while True:
            await asyncio.sleep(10)

    async def stop(self) -> None:
        """
        stop the puppet
        :return:
        """

    async def contact_list(self) -> List[str]:
        """
        get all contact list
        """

    async def tag_contact_delete(self, tag_id: str) -> None:
        """
        :return:
        """

    async def tag_favorite_delete(self, tag_id: str) -> None:
        """
        delete favorite tag from favorite
        """

    async def tag_contact_add(self, tag_id: str, contact_id: str) -> None:
        """
        add tag to contact
        :param tag_id:
        :param contact_id:
        :return:
        """

    async def tag_favorite_add(self, tag_id: str, contact_id: str) -> None:
        """
        add tag to favorite
        """

    async def tag_contact_remove(self, tag_id: str, contact_id: str) -> None:
        """
        :param tag_id:
        :param contact_id:
        :return:
        """

    async def tag_contact_list(self, contact_id: Optional[str] = None) -> List[str]:
        """
        get tag list
        :return: tag_list
        """

    async def message_send_text(self, conversation_id: str, message: str,
                                mention_ids: List[str] = None) -> str:
        """
        send text message
        :param mention_ids:
        :param conversation_id: person contact_id or room_id
        :param message: message content
        :return: message_id
        """

    async def message_send_contact(
            self,
            contact_id: str,
            conversation_id: str) -> str:
        """
        send contact message
        :param conversation_id:
        :param contact_id: person contact_id
        """

    async def message_send_file(self, conversation_id: str, file: FileBox
                                ) -> str:
        """
        send file
        :param conversation_id:
        :param file: filebox instance
        """

    async def message_send_url(
            self,
            conversation_id: str,
            url: str) -> str:
        """
        send url
        :param conversation_id:
        :param url: UrlLink instance
        """

    async def message_send_mini_program(
            self,
            conversation_id: str,
            mini_program: MiniProgramPayload) -> str:
        """
        send mini_program message
        :param conversation_id:
        :param mini_program:
        :return:
        """

    async def message_search(
            self,
            query: Optional[MessageQueryFilter] = None) -> List[str]:
        """
        search message
        """

    async def message_recall(self, message_id: str) -> bool:
        """
        send recall message
        :param message_id:
        :return:
        """

    async def message_payload(self, message_id: str) -> MessagePayload:
        """
        get message payload
        :param message_id:
        :return:
        """

    async def message_forward(self, to_id: str, message_id: str) -> None:
        """
        forward message
        :param to_id: type of Room/Contact
        :param message_id:
        :return:
        """

    async def message_file(self, message_id: str) -> FileBox:
        """
        convert message to filebox
        :param message_id:
        :return:
        """

    async def message_contact(self, message_id: str) -> str:
        """
        get message's contact card
        :param message_id:
        :return:
        """

    async def message_url(self, message_id: str) -> UrlLinkPayload:
        """
        get message' url payload
        :param message_id:
        :return:
        """

    async def message_mini_program(self, message_id: str) -> MiniProgramPayload:
        """
        get message payload
        :param message_id:
        :return:
        """

    async def contact_alias(self, contact_id: str,
                            alias: Optional[str] = None) -> str:
        """
        set contact alias
        """

    async def contact_payload_dirty(self, contact_id: str) -> None:
        """
        refresh contact payload
        """

    async def contact_payload(self, contact_id: str) -> ContactPayload:
        """
        get contact payload
        """

    async def contact_avatar(self, contact_id: str,
                             file_box: Optional[FileBox] = None
                             ) -> FileBox:
        """
        get the avatar of the account
        """

    async def contact_tag_ids(self, contact_id: str) -> List[str]:
        """
        get tag_ids of the account
        """

    def self_id(self) -> str:
        """
        get self_id
        """

    async def friendship_search(self, weixin: Optional[str] = None,
                                phone: Optional[str] = None
                                ) -> Optional[str]:
        """
        search friend by query
        :params:
        :return:
        """

    async def friendship_add(self, contact_id: str, hello: str) -> None:
        """
        add friendship with hello
        """

    async def friendship_payload(
            self,
            friendship_id: str,
            payload: Optional[FriendshipPayload] = None) -> FriendshipPayload:
        """
        load friendship payload
        """

    async def friendship_accept(self, friendship_id: str) -> None:
        """
        accept friendship
        """

    async def room_list(self) -> List[str]:
        """
        get room list
        :return:
        """

    async def room_create(self, contact_ids: List[str], topic: str = None
                          ) -> str:
        """
        create room with contact_ids and topic
        """

    async def room_search(self, query: RoomQueryFilter = None) -> List[str]:
        """
        search room by query filter
        """

    async def room_invitation_payload(self, room_invitation_id: str,
                                      payload: Optional[RoomInvitationPayload]
                                      = None) -> RoomInvitationPayload:
        """
        get room invitation payload
        """

    async def room_invitation_accept(self, room_invitation_id: str) -> None:
        """
        get room invitation payload
        """

    async def contact_self_qr_code(self) -> str:
        """
        get login qrcode
        """

    async def contact_self_name(self, name: str) -> None:
        """
        change bot contact name
        """

    async def contact_signature(self, signature: str) -> None:
        """
        change signature
        """

    async def room_payload(self, room_id: str) -> RoomPayload:
        """
        get room payload
        """

    async def room_members(self, room_id: str) -> List[str]:
        """
        get room members
        """

    async def room_add(self, room_id: str, contact_id: str) -> None:
        """
        add contact to a room
        """

    async def room_delete(self, room_id: str, contact_id: str) -> None:
        """
        delete room
        """

    async def room_quit(self, room_id: str) -> None:
        """
        quit from
        """

    async def room_topic(self, room_id: str, new_topic: str) -> None:
        """
        set room topic
        """

    async def room_announce(
            self,
            room_id: str,
            announcement: str = None) -> str:
        """
        set/get room announcement
        """

    async def room_qr_code(self, room_id: str) -> str:
        """
        get room_qrcode
        """

    async def room_member_payload(
            self,
            room_id: str,
            contact_id: str) -> RoomMemberPayload:
        """
        get room member payload
        """

    async def room_avatar(self, room_id: str) -> FileBox:
        """
        get the avatar of the room
        """

    async def logout(self) -> None:
        """
        logout the account
        :return:
        """

    async def login(self, user_id: str) -> None:
        """
        login the account
        :return:
        """

    async def dirty_payload(self, payload_type: PayloadType, payload_id: str) -> None:
        """
        mark the payload dirty status, and remove
        """



"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2020-now @ Copyright Wechaty

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations
from enum import Enum
import os
from typing import Any, Optional, List, Dict, Union
from dataclasses import dataclass

from quart import jsonify, Response
from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore
from wechaty.config import config


@dataclass
class NavMetadata:
    """nav metadata"""
    view_url: Optional[str] = None
    author: Optional[str] = None    # name of author
    avatar: Optional[str] = None    # avatar of author
    author_link: Optional[str] = None    # introduction link of author
    icon: Optional[str] = None    # avatar of author


@dataclass
class NavDTO:
    """the data transfer object of plugin list"""
    name: str                       # name of plugin
    status: int                     # status of plugin: 0 / 1

    view_url: Optional[str] = None
    author: Optional[str] = None    # name of author
    avatar: Optional[str] = None    # avatar of author
    author_link: Optional[str] = None    # introduction link of author
    icon: Optional[str] = None    # avatar of author

    def update_metadata(self, nav_metadata: NavMetadata) -> None:
        """update the field with nav data
        """
        self.author = nav_metadata.author
        self.author_link = nav_metadata.author_link
        self.avatar = nav_metadata.avatar
        self.icon = nav_metadata.icon
        self.view_url = nav_metadata.view_url


def success(data: Any) -> Response:
    """make the success response with data

    Args:
        data (dict): the data of response
    """
    return jsonify(dict(
        code=200,
        data=data
    ))


def error(msg: str) -> Response:
    """make the error response with msg

    Args:
        msg (str): the error msg string of data
    """
    return jsonify(dict(
        code=500,
        msg=msg
    ))


@dataclass
class WechatyPluginOptions:
    """options for wechaty plugin"""
    name: Optional[str] = None
    metadata: Optional[dict] = None


@dataclass
class WechatySchedulerOptions:
    """options for wechaty scheduler"""
    job_store: Union[str, SQLAlchemyJobStore] = f'sqlite:///{config.cache_dir}/job.db'
    job_store_alias: str = 'wechaty-scheduler'


class PluginStatus(Enum):
    """plugin running status"""
    Running = 0
    Stopped = 1


class StaticFileCacher:
    """cache the static file to avoid time-consuming finding and loading
    """
    def __init__(self, cache_dirs: Optional[List[str]] = None) -> None:
        self.file_maps: Dict[str, str] = {}

        self.cache_dirs = cache_dirs or []

    def add_dir(self, static_file_dir: Optional[str]) -> None:
        """add the static file dir

        Args:
            static_file_dir (str): the path of the static file
        """
        if not static_file_dir:
            return
        self.cache_dirs.append(static_file_dir)

    def _find_file_path_recursive(self, base_dir: str, name: str) -> Optional[str]:
        """find the file based on the file-name which will & should be union

        Args:
            base_dir (str): the root dir of static files for the plugin
            name (str): the union name of static file

        Returns:
            Optional[str]: the target static file path
        """
        if not os.path.exists(base_dir) or os.path.isfile(base_dir):
            return None

        for file_name in os.listdir(base_dir):
            if file_name == name:
                return os.path.join(base_dir, file_name)
            file_path = os.path.join(base_dir, file_name)

            target_path = self._find_file_path_recursive(file_path, name)
            if target_path:
                return target_path

        return None

    def find_file_path(self, name: str) -> Optional[str]:
        """find the file based on the file-name which will & should be union

        Args:
            name (str): the union name of static file

        Returns:
            Optional[str]: the path of the static file
        """
        if name in self.file_maps:
            return self.file_maps[name]

        for cache_dir in self.cache_dirs:
            file_path = self._find_file_path_recursive(cache_dir, name)
            if file_path:
                return file_path
        return None



"""link helper function for fetching the friendly meta data from url link"""
from __future__ import annotations
from typing import Any, Dict
import requests
from opengraph_py3 import OpenGraph # type: ignore


def get_url_metadata(url: str) -> Dict[str, Any]:
    """get open graph meta data open open graph protocol

    The Open Graph Protocol: https://ogp.me/

    Args:
        url (str): the url of link

    Returns:
        Dict[str, Any]: the meta data
    """
    return OpenGraph(url=url)


def fetch_github_user_avatar_url(name: str) -> str:
    """fetch_github_user_avatar_url

    refer to: https://docs.github.com/en/rest/users/users#get-a-user

    Args:
        name (str): the name of github user

    Returns:
        str: the avatar url of github user
    """
    source_avatar_url = f'https://api.github.com/users/{name}'

    header = {
        "accept": "application/vnd.github.v3+jso"
    }
    response = requests.get(source_avatar_url, headers=header)
    data = response.json()
    return data.get('avatar_url', None)


"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2020-now @ Copyright Wechaty

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

from typing import (
    Callable,
    Any,
    Dict,
    Iterable,
    Iterator,
    Optional,
    Type,
    TypeVar,
    MutableMapping,
    Generic
)

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class HookDict(MutableMapping[_KT, _VT], Generic[_KT, _VT]):
    """add hooks to the dict data
    """
    def __init__(
        self,
        data: Dict[_KT, _VT],
        set_item_hooks: Optional[Callable[[_KT, _VT, Dict[_KT, _VT]], None]] = None,
        default_type: Type = int,
    ):
        self._dict_data = data
        self._set_item_hooks = set_item_hooks
        self.default_type = default_type

    def __setitem__(self, key: _KT, value: Any) -> None:
        """triggered by `data[key] = value`"""
        self._dict_data[key] = value
        if self._set_item_hooks is not None:
            self._set_item_hooks(key, value, self._dict_data)

    def __getitem__(self, __k: _KT) -> _VT:
        """get item data"""
        return self._dict_data.get(__k, self.default_type())
    
    def __delitem__(self, __v: _KT) -> None:
        """del item by key"""
        del self._dict_data[__v]
    
    def __iter__(self) -> Iterator[_KT]:
        """get the iterable data of data"""
        return iter(self._dict_data)
    
    def __len__(self) -> int:
        """get the length of dict data"""
        return len(self._dict_data)

    def to_dict(self) -> dict:
        """get the source dict data"""
        return self._dict_data
